# -*- coding: utf-8 -*-
# Copyright (C) 2019 NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import tempfile
from unittest.mock import patch

from tornado.httpclient import HTTPResponse
from tornado.testing import AsyncHTTPTestCase
from tornado.web import Application

from cylc.uiserver.main import *


class MainHandlerTest(AsyncHTTPTestCase):
    """Test for the Main handler"""
    def get_app(self) -> Application:
        self.tempdir = tempfile.mkdtemp(suffix='mainhandlertest')
        return MyApplication(
            handlers=[
                ('/', MainHandler, {"path": self.tempdir})
            ]
        )

    def test_jupyterhub_version_returned(self):
        with patch.object(MainHandler, 'get_current_user') as mocked:
            mocked.return_value = {'name': 'yossarian'}
            with open(os.path.join(self.tempdir, "index.html"), "w+") as nf:
                nf.write("TESTING!")
                nf.flush()
                response = self.fetch("/")  # type: HTTPResponse
                assert response.body == b"TESTING!"
                assert "X-JupyterHub-Version" in response.headers

    def tearDown(self) -> None:
        if self.tempdir:
            shutil.rmtree(self.tempdir, ignore_errors=True)


class UserProfileHandlerTest(AsyncHTTPTestCase):
    """Test for UserProfile handler"""
    def get_app(self) -> Application:
        return MyApplication(
            handlers=[
                ('/userprofile', UserProfileHandler)
            ]
        )

    def test_user_profile_handler_cors_headers(self):
        with patch.object(UserProfileHandler, 'get_current_user') as mocked:
            mocked.return_value = {'name': 'yossarian'}
            response = self.fetch("/userprofile")  # type: HTTPResponse
            assert "Access-Control-Allow-Origin" in response.headers
            assert "Access-Control-Allow-Headers" in response.headers
            assert "Access-Control-Allow-Methods" in response.headers
            assert "Content-Type" in response.headers
            assert b'yossarian' in response.body



# -*- coding: utf-8 -*-
# Copyright (C) 2019 NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from cylc.uiserver.main import *


def test_my_application():
    """Test creating the Tornado app."""
    my_application = MyApplication(handlers=[])
    assert not my_application.is_closing
    my_application.signal_handler(None, None)
    assert my_application.is_closing


def test_cylcuiserver_absolute_path():
    """Test a Cylc UI server created with absolute path for static assets."""
    cylc_uiserver = CylcUIServer(8000, '/static/path', '/users/test')
    assert cylc_uiserver._port == 8000
    assert cylc_uiserver._static == '/static/path'
    assert cylc_uiserver._jupyter_hub_service_prefix == '/users/test'


def test_cylcuiserver_relative_path():
    """Test a Cylc UI server created with relative path for static assets."""
    cylc_uiserver = CylcUIServer(8000, './', '/users/test')
    assert cylc_uiserver._port == 8000
    print(cylc_uiserver._static)
    # the code is using the directory relative to the main script ATM
    assert cylc_uiserver._static.endswith('cylc/uiserver')
    assert cylc_uiserver._jupyter_hub_service_prefix == '/users/test'



# -*- coding: utf-8 -*-
# Copyright (C) 2019 NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Manage Workflow services.

Includes:
    - Spawning (#TODO)
    - Scans/updates
    - ZeroMQ REQ/RES client initiation and checks
    - ?

"""
import socket
import asyncio
import logging

from contextlib import suppress

from cylc.flow import flags
from cylc.flow.exceptions import ClientError, ClientTimeout
from cylc.flow.hostuserutil import is_remote_host, get_host_ip_by_name
from cylc.flow.network.client import SuiteRuntimeClient
from cylc.flow.network.scan import (
    get_scan_items_from_fs, re_compile_filters, MSG_TIMEOUT)
from cylc.flow.ws_data_mgr import ID_DELIM

logger = logging.getLogger(__name__)
CLIENT_TIMEOUT = 2.0


async def workflow_request(client, command, args=None,
                           timeout=None, context=None):
    """Workflow request command."""
    if context is None:
        context = command
    try:
        result = await client.async_request(command, args, timeout)
        return (context, result)
    except ClientTimeout as exc:
        logger.exception(exc)
        return (context, MSG_TIMEOUT)
    except ClientError as exc:
        logger.exception(exc)
        return (context, None)


async def est_workflow(reg, host, port, timeout=None):
    """Establish communication with workflow, instantiating REQ client."""
    if is_remote_host(host):
        try:
            host = get_host_ip_by_name(host)  # IP reduces DNS traffic
        except socket.error as exc:
            if flags.debug:
                raise
            logger.error("ERROR: %s: %s\n" % (exc, host))
            return (reg, host, port, None)

    # NOTE: Connect to the suite by host:port. This way the
    #       SuiteRuntimeClient will not attempt to check the contact file
    #       which would be unnecessary as we have already done so.
    # NOTE: This part of the scan *is* IO blocking.
    client = SuiteRuntimeClient(reg, host=host, port=port, timeout=timeout)
    context, result = await workflow_request(client, 'identify')
    return (reg, host, port, client, result)


class WorkflowsManager(object):

    def __init__(self):
        self.workflows = {}

    def spawn_workflow(self):
        # TODO - Spawn workflows
        pass

    async def gather_workflows(self):
        scanflows = {}
        cre_owner, cre_name = re_compile_filters(None, ['.*'])
        scan_args = (
            (reg, host, port, CLIENT_TIMEOUT)
            for reg, host, port in
            get_scan_items_from_fs(cre_owner, cre_name))
        gathers = ()
        for arg in scan_args:
            gathers += (est_workflow(*arg),)
        items = await asyncio.gather(*gathers)
        for reg, host, port, client, info in items:
            if info is not None and info != MSG_TIMEOUT:
                owner = info['owner']
                scanflows[f"{owner}{ID_DELIM}{reg}"] = {
                    'name': info['name'],
                    'owner': owner,
                    'host': host,
                    'port': port,
                    'version': info['version'],
                    'req_client': client,
                }

        # Check existing against scan
        for w_id, info in list(self.workflows.items()):
            if w_id in scanflows:
                if (info['host'] == scanflows[w_id]['host'] and
                        info['port'] == scanflows[w_id]['port']):
                    client = scanflows[w_id]['req_client']
                    with suppress(IOError):
                        client.socket.close()
                    scanflows.pop(w_id)
                continue
            client = self.workflows[w_id]['req_client']
            with suppress(IOError):
                client.socket.close()
            self.workflows.pop(w_id)

        # update with new
        self.workflows.update(scanflows)

    async def multi_request(self, command, workflows, args=None,
                            multi_args=None, timeout=None):
        """Send requests to multiple workflows."""
        if args is None:
            args = {}
        if multi_args is None:
            multi_args = {}
        req_args = {}
        for w_id in workflows:
            cmd_args = multi_args.get(w_id, args)
            req_args[w_id] = (
                self.workflows[w_id]['req_client'],
                command,
                cmd_args,
                timeout,
            )
        gathers = ()
        for info, request_args in req_args.items():
            gathers += (workflow_request(context=info, *request_args),)
        results = await asyncio.gather(*gathers)
        res = []
        for key, val in results:
            res.extend([
                msg_core
                for msg_core in list(val.values())[0].get('result')
                if isinstance(val, dict) and list(val.values())])
        return res



#!/usr/bin/env python3

# -*- coding: utf-8 -*-
# Copyright (C) 2019 NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import logging
import os
import signal

from cylc.flow.network.schema import schema
from logging.config import dictConfig
from os.path import join, abspath, dirname
from tornado import web, ioloop

from jupyterhub.services.auth import HubOAuthCallbackHandler
from jupyterhub.utils import url_path_join
from .data_mgr import DataManager
from .handlers import *
from .resolvers import Resolvers
from .workflows_mgr import WorkflowsManager

logger = logging.getLogger(__name__)


class MyApplication(web.Application):
    is_closing = False

    def signal_handler(self, signum, frame):
        logger.info('exiting...')
        self.is_closing = True

    def try_exit(self):
        if self.is_closing:
            # clean up here
            ioloop.IOLoop.instance().stop()
            logger.info('exit success')


class CylcUIServer(object):

    def __init__(self, port, static, jupyter_hub_service_prefix):
        self._port = port
        if os.path.isabs(static):
            self._static = static
        else:
            script_dir = os.path.dirname(__file__)
            self._static = os.path.abspath(os.path.join(script_dir, static))
        self._jupyter_hub_service_prefix = jupyter_hub_service_prefix
        self.ws_mgr = WorkflowsManager()
        self.data_mgr = DataManager(self.ws_mgr)
        self.resolvers = Resolvers(
            self.data_mgr.data,
            ws_mgr=self.ws_mgr)

    def _make_app(self, debug: bool):
        """Crete a Tornado web application.

        Args:
            debug (bool): flag to set debugging in the Tornado application
        """
        logger.info(self._static)
        return MyApplication(
            static_path=self._static,
            debug=debug,
            handlers=[
                (rf"{self._jupyter_hub_service_prefix}(.*.(css|js))",
                 web.StaticFileHandler, {"path": self._static}),
                (rf"{self._jupyter_hub_service_prefix}((fonts|img)/.*)",
                 web.StaticFileHandler, {"path": self._static}),
                (rf"{self._jupyter_hub_service_prefix}(favicon.png)",
                 web.StaticFileHandler, {"path": self._static}),

                (r"/(.*.(css|js))",
                 web.StaticFileHandler, {"path": self._static}),
                (r"/((fonts|img)/.*)",
                 web.StaticFileHandler, {"path": self._static}),
                (r"/(favicon.png)",
                 web.StaticFileHandler, {"path": self._static}),

                (url_path_join(
                    self._jupyter_hub_service_prefix, 'oauth_callback'),
                    HubOAuthCallbackHandler),
                (url_path_join(
                    self._jupyter_hub_service_prefix, 'userprofile'),
                    UserProfileHandler),
                (url_path_join(
                    self._jupyter_hub_service_prefix, 'suites'),
                    CylcScanHandler),

                (rf"{self._jupyter_hub_service_prefix}?",
                    MainHandler, {"path": self._static}),

                # graphql
                (url_path_join(self._jupyter_hub_service_prefix,
                               '/graphql'),
                    UIServerGraphQLHandler,
                    dict(schema=schema, resolvers=self.resolvers)),
                (url_path_join(self._jupyter_hub_service_prefix,
                               '/graphql/batch'),
                    UIServerGraphQLHandler,
                    dict(schema=schema, resolvers=self.resolvers, batch=True)),
                (url_path_join(self._jupyter_hub_service_prefix,
                               '/graphql/graphiql'),
                    UIServerGraphQLHandler,
                    dict(schema=schema, resolvers=self.resolvers,
                         graphiql=True)),
            ],
            # FIXME: decide (and document) whether cookies will be permanent
            # after server restart.
            cookie_secret="cylc-secret-cookie"
        )

    def start(self, debug: bool):
        app = self._make_app(debug)
        signal.signal(signal.SIGINT, app.signal_handler)
        app.listen(self._port)
        ioloop.PeriodicCallback(app.try_exit, 100).start()
        # Discover workflows on initial start up.
        ioloop.IOLoop.current().add_callback(self.ws_mgr.gather_workflows)
        # If the client is already established it's not overridden,
        # so the following callbacks can happen at the same time.
        ioloop.PeriodicCallback(self.ws_mgr.gather_workflows, 10000).start()
        ioloop.PeriodicCallback(
            self.data_mgr.entire_workflow_update, 5000).start()
        try:
            ioloop.IOLoop.current().start()
        except KeyboardInterrupt:
            ioloop.IOLoop.instance().stop()


def main():
    parser = argparse.ArgumentParser(
        description="Start Cylc UI"
    )
    parser.add_argument('-p', '--port', action="store", dest="port", type=int,
                        default=8888)
    parser.add_argument('-s', '--static', action="store", dest="static",
                        required=True)
    parser.add_argument('--debug', action="store_true", dest="debug",
                        default=False)
    here = abspath(dirname(__file__))
    parser.add_argument('--logging-config', type=argparse.FileType('r'),
                        help='path to logging configuration file',
                        action="store", dest="logging_config",
                        default=join(here, 'logging_config.json'))
    args = parser.parse_known_args()[0]

    # args.logging_config will be a io.TextIOWrapper resource
    with args.logging_config as logging_config_json:
        config = json.load(logging_config_json)
        dictConfig(config["logging"])

    jupyterhub_service_prefix = os.environ.get(
        'JUPYTERHUB_SERVICE_PREFIX', '/')
    logger.info(f"JupyterHub Service Prefix: {jupyterhub_service_prefix}")
    ui_server = CylcUIServer(
        port=args.port,
        static=args.static,
        jupyter_hub_service_prefix=jupyterhub_service_prefix
    )
    logger.info(f"Listening on {args.port} and serving static content from "
                f"{args.static}")

    logger.info("Starting Cylc UI")
    ui_server.start(args.debug)


if __name__ == "__main__":
    main()


__all__ = ['MainHandler', 'UserProfileHandler', 'MyApplication',
           'CylcUIServer', 'main']



# -*- coding: utf-8 -*-
# Copyright (C) 2019 NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import re
from subprocess import Popen, PIPE
from typing import List, Union

from jupyterhub import __version__ as jupyterhub_version
from jupyterhub.services.auth import HubOAuthenticated
from tornado import web
from graphene_tornado.tornado_graphql_handler import TornadoGraphQLHandler
from graphql import get_default_backend


class BaseHandler(web.RequestHandler):

    def set_default_headers(self) -> None:
        self.set_header("X-JupyterHub-Version", jupyterhub_version)
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "x-requested-with")
        self.set_header('Access-Control-Allow-Methods', 'POST, GET, OPTIONS')


class APIHandler(BaseHandler):

    def set_default_headers(self) -> None:
        super().set_default_headers()
        self.set_header("Content-Type", 'application/json')


class MainHandler(HubOAuthenticated, BaseHandler):

    # hub_users = ["kinow"]
    # hub_groups = []
    # allow_admin = True

    def initialize(self, path):
        self._static = path

    @web.addslash
    @web.authenticated
    def get(self):
        """Render the UI prototype."""
        index = os.path.join(self._static, "index.html")
        self.write(open(index).read())


class UserProfileHandler(HubOAuthenticated, APIHandler):

    def set_default_headers(self) -> None:
        super().set_default_headers()
        self.set_header("Content-Type", 'application/json')

    @web.authenticated
    def get(self):
        self.write(json.dumps(self.get_current_user()))


class CylcScanHandler(HubOAuthenticated, APIHandler):

    def _parse_suite_line(self, suite_line: bytes) -> Union[dict, None]:
        if suite_line:
            parts = re.split(r"[ :@]", suite_line.decode())
            if len(parts) == 4:
                return {
                    "name": parts[0],
                    "user": parts[1],
                    "host": parts[2],
                    "port": int(parts[3])
                }
        return None

    def _get_suites(self, suite_lines: List[bytes]) -> List[dict]:
        suites = []
        for suite_line in suite_lines:
            suite = self._parse_suite_line(suite_line)
            if suite:
                suites.append(suite)
        return suites

    @web.authenticated
    def get(self):
        cylc_scan_proc = Popen(
            "cylc scan --color=never", shell=True, stdout=PIPE)
        cylc_scan_out = cylc_scan_proc.communicate()[0]

        suite_lines = cylc_scan_out.splitlines()
        suites = self._get_suites(suite_lines)
        self.write(json.dumps(suites))


# This is needed in order to pass the server context in addition to existing.
# It's possible to just overwrite TornadoGraphQLHandler.context but we would
# somehow need to pass the request info (headers, username ...etc) in also
class UIServerGraphQLHandler(HubOAuthenticated, TornadoGraphQLHandler):

    # Declare extra attributes
    resolvers = None

    def initialize(self, schema=None, executor=None, middleware=None,
                   root_value=None, graphiql=False, pretty=False,
                   batch=False, backend=None, **kwargs):
        super(TornadoGraphQLHandler, self).initialize()

        self.schema = schema
        if middleware is not None:
            self.middleware = list(self.instantiate_middleware(middleware))
        self.executor = executor
        self.root_value = root_value
        self.pretty = pretty
        self.graphiql = graphiql
        self.batch = batch
        self.backend = backend or get_default_backend()
        # Set extra attributes
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    @property
    def context(self):
        wider_context = {
            'graphql_params': self.graphql_params,
            'request': self.request,
            'resolvers': self.resolvers,
        }
        return wider_context

    @web.authenticated
    def prepare(self):
        super().prepare()


__all__ = [
    "MainHandler",
    "UserProfileHandler",
    "CylcScanHandler",
    "UIServerGraphQLHandler"
]



# -*- coding: utf-8 -*-
# Copyright (C) 2019 NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Create and update the data structure for all workflow services."""

import asyncio
import logging

from cylc.flow.exceptions import ClientError, ClientTimeout
from cylc.flow.network.server import PB_METHOD_MAP
from cylc.flow.network.scan import MSG_TIMEOUT
from cylc.flow.ws_data_mgr import (
    EDGES, FAMILIES, FAMILY_PROXIES, JOBS, TASKS, TASK_PROXIES, WORKFLOW
)

logger = logging.getLogger(__name__)


async def get_workflow_data(w_id, client, method):
    """Call WS endpoint for entire workflow protobuf message."""
    # Use already established client
    try:
        pb_msg = await client.async_request(method)
    except ClientTimeout as exc:
        logger.exception(exc)
        return (w_id, MSG_TIMEOUT)
    except ClientError as exc:
        logger.exception(exc)
        return (w_id, None)
    else:
        ws_data = PB_METHOD_MAP[method]()
        ws_data.ParseFromString(pb_msg)
    return (w_id, ws_data)


class DataManager:
    """Manage the local data-store acquisition/updates from all workflows."""

    def __init__(self, ws_mgr):
        self.ws_mgr = ws_mgr
        self.data = {}

    # Data syncing
    async def entire_workflow_update(self, ids=None):
        """Update all data of workflow(s) from associated WS."""
        if ids is None:
            ids = []

        # Prune old data
        for w_id in list(self.data):
            if w_id not in self.ws_mgr.workflows:
                del self.data[w_id]

        # Fetch new data
        ws_args = (
            (w_id, info['req_client'], 'pb_entire_workflow')
            for w_id, info in self.ws_mgr.workflows.items())
        gathers = ()
        for args in ws_args:
            if not ids or args[0] in ids:
                gathers += (get_workflow_data(*args),)
        items = await asyncio.gather(*gathers)
        new_data = {}
        for w_id, result in items:
            if result is not None and result != MSG_TIMEOUT:
                new_data[w_id] = {
                    EDGES: {e.id: e for e in getattr(result, EDGES)},
                    FAMILIES: {f.id: f for f in getattr(result, FAMILIES)},
                    FAMILY_PROXIES: {
                        n.id: n
                        for n in getattr(result, FAMILY_PROXIES)},
                    JOBS: {j.id: j for j in getattr(result, JOBS)},
                    TASKS: {t.id: t for t in getattr(result, TASKS)},
                    TASK_PROXIES: {
                        n.id: n
                        for n in getattr(result, TASK_PROXIES)},
                    WORKFLOW: getattr(result, WORKFLOW),
                }

        self.data.update(new_data)



# -*- coding: utf-8 -*-
# Copyright (C) 2019 NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""GraphQL resolvers for use in data accessing and mutation of workflows."""

from cylc.flow.network.resolvers import BaseResolvers
from cylc.flow.ws_data_mgr import WORKFLOW


class Resolvers(BaseResolvers):
    """UI Server context GraphQL query and mutation resolvers."""

    ws_mgr = None

    def __init__(self, data, **kwargs):
        super().__init__(data)

        # Set extra attributes
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    # Mutations
    async def mutator(self, info, *m_args):
        """Mutate workflow."""
        _, w_args, _ = m_args
        w_ids = [
            flow[WORKFLOW].id
            for flow in await self.get_workflows_data(w_args)]
        # Pass the request to the workflow GraphQL endpoints
        req_str, variables, _, _ = info.context.get('graphql_params')
        graphql_args = {
            'request_string': req_str,
            'variables': variables,
        }
        return self.ws_mgr.multi_request('graphql', w_ids, graphql_args)

    async def nodes_mutator(self, info, *m_args):
        """Mutate node items of associated workflows."""
        _, _, w_args, _ = m_args
        w_ids = [
            flow[WORKFLOW].id
            for flow in await self.get_workflows_data(w_args)]
        if not w_ids:
            return 'Error: No matching Workflow'
        # Pass the multi-node request to the workflow GraphQL endpoints
        req_str, variables, _, _ = info.context.get('graphql_params')
        graphql_args = {
            'request_string': req_str,
            'variables': variables,
        }
        multi_args = {w_id: graphql_args for w_id in w_ids}
        return self.ws_mgr.multi_request(
            'graphql', w_ids, multi_args=multi_args)



# -*- coding: utf-8 -*-
# Copyright (C) 2019 NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import codecs
import re
from os.path import join, abspath, dirname

from setuptools import setup, find_namespace_packages

here = abspath(dirname(__file__))


def read(*parts):
    with codecs.open(join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


install_requires = [
    'jupyterhub==1.0.*',
    'tornado==6.0.*',
    'graphene-tornado==2.1.*',
    'cylc-flow==8.0a1'
]

setup_requires = [
    'pytest-runner==4.4.*'
]

tests_require = [
    'pytest==4.4.*',
    'coverage==4.5.*',
    'pytest-cov==2.6.*'
]

extras_require = {
    'tests': tests_require,
    'all': install_requires + tests_require
}

setup(
    version=find_version("cylc", "uiserver", "__init__.py"),
    long_description=read("README.md"),
    long_description_content_type='text/markdown',
    packages=find_namespace_packages(include=["cylc.*"]),
    package_data={
        'cylc.uiserver': [
            'logging_config.json'
        ]
    },
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    extras_require=extras_require,
    entry_points={
        'console_scripts': [
            'cylc-uiserver=cylc.uiserver.main:main'
        ]
    }
)
